/***********************************************************************

AGB-QEDS v0.05 BETA by Herman Haverkort, 10 May 2022

************************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of AGB-QEDS may be available from 
http://herman.haverkort.net.


************************************************************************

WHAT IS AGB-QEDS?

AGB-QEDS is an implementation of a Quad Edge Data Structure (QEDS) that 
stores a connected planar graph. It is based on the design by Guibas and 
Stolfi [ACM Trans. Graph. 4(2):74-123, 1985], extended in three ways:
- the splice-operation (the basic operation that sets all the pointers) 
  is encapsulated in operations with names that describe the intended 
  effects, and which verify (if desired for debugging purposes) the 
  preconditions for obtaining those effects;
- the structure stores pointers to face and vertex information;
- several useful higher-level operations are provided, in particular 
  for the implementation of the divide-and-conquer and randomised 
  incremental construction algorithms for Voronoi Diagrams and Delaunay
  triangulations as described in Klein, Driemel and Haverkort:
  Algorithmische Geometrie, 3rd edition, 2022.


***********************************************************************/

#ifndef AGB_QEDS_INCLUDED
#define AGB_QEDS_INCLUDED

/**********************************************************************/

// CONFIGURATION

// It is possible to verify the preconditions for certain operations 
// and the consistency of vertex and face information at run time. 
// However, the verification steps traverse all incident edges of the 
// vertices/faces involved in such operations. This may turn constant-
// time operations into linear-time operations, in the worst case 
// slowing down applications considerably.

// To verify preconditions, uncomment the following line:

// #define SAFEQEDS


// To verify that edges incident on the same vertex always have the same 
// origin pointer, uncomment the following line:

// #define CONSISTENTQEDS

// Note: 
// * This also ensures consistent face information (as faces are 
//   vertices of the dual). 
// * It is still allowed that multiple vertices/faces have the same 
//   vertex/face info. This is in particular the case if no vertex/face
//   pointers are ever set (so that they are all NULL), or after a vertex 
//   or face is split: both parts will have the same label.
// * The verification step will stop any attempts to merge vertices or 
//   faces with different vertex/face pointers. Before merging, one has 
//   to relabel one of the vertices/faces so that both operands have 
//   the same label.


// libraries used:
#include <cassert>
#include <algorithm>
using std::swap;
#include <stack>
using std::stack;
#include <vector>
#include <ostream>
using std::ostream;


/***********************************************************************

NOTATION

Let e = (p,q) be an edge of the graph, where L is the face to its left
and R is the face to its right. The QEDS then distinguishes between 
four related directed edges:

e           the edge p->q
rtol(e)     the dual edge in the direction R->L
rev(e)      the edge p->q in the opposite direction, that is, q->p
ltor(e)     the dual edge in the opposite direction, that is, L->R

Following DCEL terminology, we call these directed edges "half-edges".
The QEDS explicitly stores for each half-edge e:

origin(e)   a pointer to the starting vertex of e
sccorgn(e)  the successor in the counterclockwise order among the half-
            edges with the same origin

(This differs from the presentation in Klein et al.: there, e stores the 
pointers that are here associated with ltor(e). In this implementation,
we follow Guibas and Stolfi more closely.) Using these pointers, the QEDS 
also provides indirect access to the following information about each 
half-edge e:

destin(e)   = origin(rev(e)): the destination vertex of e
leftface(e) = origin(ltor(e)): a pointer to the face to the left of e
rightface(e)= origin(rtol(e)): a pointer to the face to the right of e
pccorgn(e)  = rtol(sccorgn(rtol(e))): the predecessor in the counter-
              clockwise order among the half-edges with the same origin
sccleft(e)  = rtol(sccorgn(ltor(e))): the successor in the counter-
              clockwise order along the boundary of the face to the left
pccleft(e)  = rev(sccorgn(e)): the predecessor in the counterclockwise
              order along the boundary of the face to the left
scright(e)  = sccorgn(rev(e))): the successor in the clockwise order 
              along the boundary of the face to the right
pcright(e)  = ltor(sccorgn(rtol(e)): the predecessor in the clockwise
              order along the boundary of the face to the right
partner(e)  = rtol(sccorgn(e)) = pccorgn(ltor(e))


***********************************************************************/

namespace AGBQEDS {

/**********************************************************************/

// HalfEdge HEADER

/**********************************************************************/

template <class V, class F>
class HalfEdge
{
  // The HalfEdge class is templated with the vertex type V and the face 
  // type F; thus the compiler can enforce that we do not confuse primal
  // and dual edges (provided vertices and faces have different types).


  //////////////////////////////////////////////////////////////////////
  // fundamental members                                              //
  //////////////////////////////////////////////////////////////////////
    private:

  // the index of the half-edge in its quadedge (pack of four halfedges)
  short int halfEdgeIndex; 

  // the data:
  void* pOrigin;
  void* pSuccessor;  

  // markings:
  bool marking;
  void mark();
  void unmark();
  bool marked() const;

    public:
  // report number of edges alive (to detect memory leaks)
  static int totalNrAlive();

  // friends (classes whose methods may access the private member 
  // variables and methods of HalfEdge):
  friend class HalfEdge<F,V>;  // (the dual type of HalfEdge)

  
  //////////////////////////////////////////////////////////////////////
  // public access to the half-edge and its neighbourhood:            //
  //////////////////////////////////////////////////////////////////////

    public:
  HalfEdge<V,F>* rev();
  HalfEdge<F,V>* rtol();
  HalfEdge<F,V>* ltor();
  V* origin();
  V* destin();
  F* leftFace();
  F* rightFace();
  HalfEdge<V,F>* sccorgn();
  HalfEdge<V,F>* pccorgn();
  HalfEdge<V,F>* sccleft();
  HalfEdge<V,F>* pccleft();
  HalfEdge<V,F>* scright();
  HalfEdge<V,F>* pcright();
  HalfEdge<F,V>* partner();


  //////////////////////////////////////////////////////////////////////
  // traversal operators:                                             //
  //////////////////////////////////////////////////////////////////////
    public:

  // collect all edges 
  // (one out of each primal half-edge and its reverse):
  std::vector<HalfEdge<V,F>* > getEdges();

    public:
  // collect all vertices (one outgoing edge for each vertex)
  std::vector<HalfEdge<V,F>* > getVertices();


  //////////////////////////////////////////////////////////////////////
  // counting and verification operators:                             //
  //////////////////////////////////////////////////////////////////////
    public:

  // get number of edges incident on start(e) (loops count twice):
  int nrEdgesOnOrigin();

  // get the complexity of the boundary of the face to the left of e:
  int nrEdgesOnLeftFace();

  // check whether *this and *e have the same origin. Note: this follows
  // the cycles of edge around the origin; it does not rely on the 
  // pointers to vertex information! The loop follows the cycles around
  // both origins simultaneously and stops, at the latest, when the 
  // smallest of them has been traversed completely.
  bool hasSameOrigin(HalfEdge<V,F>* e);

  // check whether *this and *e have the same face to the left. Note: 
  // this follows the cycle of edges around the face; it does not rely 
  // on pointers to face information!
  bool hasSameLeftFace(HalfEdge<V,F>* e);


  //////////////////////////////////////////////////////////////////////
  // operations to create a graph                                     //
  //////////////////////////////////////////////////////////////////////
    private:
  // private constructor and post-constructor for individual halfedges:
  HalfEdge<V,F>();
  void setData(int _halfEdgeIndex, void* _pOrigin, void* _pSuccessor);

    public:
  // create a graph that consists of an edge from origin to destination 
  // with the given face as the outer and only face:
  static HalfEdge<V,F>* createEdge(
    V* origin = NULL, V* destination = NULL, F* face = NULL);

  // create a graph that consists of a loop with the specified inner
  // face, outer face, and vertex; the half-edge that is returned has 
  // the inner face on the left:
  static HalfEdge<V,F>* createLoop(
    F* innerFace = NULL, V* vertex = NULL, F* outerFace = NULL);


  //////////////////////////////////////////////////////////////////////
  // operations to modify a graph                                     //
  //////////////////////////////////////////////////////////////////////

  // splice swaps sccorgn(e1) and sccorgn(e2), as well as 
  // sccorgn(partner(e1)) and sccorgn(partner(e2));
  // both swaps change whether the operands are in the same cycle around
  // a common origin (from yes to no or from no to yes);
  // splice is used by almost all methods that modify the graph;
  // we keep it private because it could have strange effects if used 
  // on pairs of edges it should not be used on
    private:
  static void splice(HalfEdge<V,F>* e1, HalfEdge<V,F>* e2);

  // the following four operations simply call splice, but after 
  // checking different preconditions that are neccesary for the 
  // intended effect to happen:

  // cut the common origin of e and f into two vertices, one with 
  // edges sccorgn(e),...,f and one with edges sccorgn(f),...,e;
  // thus merging the faces to the left of e and f.
  // (not to be used if it would render the graph disconnected)
    public:
  static void splitOrigin(HalfEdge<V,F>* e, HalfEdge<V,F>* f);

  // the same as above when e and f have the same face to the left;
  // this will render the graph disconnected
    public:
  static void splitGraph(HalfEdge<V,F>* e, HalfEdge<V,F>* f);

  // contract the origins of e and f into a single vertex, thus
  // cutting their common face into two faces
  static void joinOrigins(HalfEdge<V,F>* e, HalfEdge<V,F>* f);

  // contract the origins of the current edge and e into a single 
  // vertex to insert the graph with e on the outer face into the face 
  // to the left of f (or vice versa); 
  // it should not be used if e and f are in the same 
  // graph already.
  static void insertGraph(HalfEdge<V,F>* e, HalfEdge<V,F>* f);


  //////////////////////////////////////////////////////////////////////
  // operations to modify vertices and faces                          //
  //////////////////////////////////////////////////////////////////////
    public:

  // set vertex labels for all edges around the same vertex:
  void labelOrigin(V* v);

  // set face labels for all edges around the face to the left:
  void labelLeftFace(F* f);


  //////////////////////////////////////////////////////////////////////
  // operations to delete (parts of) a graph                          //
  //////////////////////////////////////////////////////////////////////

  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //   !!!   NOTE: These operations do NOT delete the vertex !!!
  //   !!!   and face information that the vertex and face   !!!
  //   !!!   pointers point to!                              !!!
  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    private:
  // private destructor for individual half-edges:
  ~HalfEdge();

  // deleteEdge disconnects the edge from the rest of the graph and 
  // destroys it. we keep it private because it may render the rest of 
  // the graph disconnected, and we want to make sure that it is 
  // used only by operations that deal with that correctly
  static void deleteEdge(HalfEdge<V,F>*& e);

    public:
  // deleteGraph deletes an entire graph
  static void deleteGraph(HalfEdge<V,F>*& e);

  // deleteSubgraph removes the component that is surrounded, in
  // clockwise order, by the edges from e to, but not including, the 
  // current edge
  void deleteSubgraph(HalfEdge<V,F>*& e);


  //////////////////////////////////////////////////////////////////////
  // advanced operations                                              //
  //////////////////////////////////////////////////////////////////////

  // cut the common origin of e and f into two vertices, one with 
  // edges sccorgn(e),...,f and one with edges sccorgn(f),...,e;
  // with a new edge connecting the two vertices; returns the new
  // edge in the direction from the origin of e to the origin of f
    public:
  static HalfEdge<V,F>* expandOrigin(HalfEdge<V,F>* e, HalfEdge<V,F>* f);

  // insert an edge between the origins of e and f, thus splitting 
  // their common face; returns the new edge in the direction from the 
  // origin of e to the origin of f
  static HalfEdge<V,F>* splitFace(HalfEdge<V,F>* e, HalfEdge<V,F>* f);

  // add a vertex on an edge; afterwards *this will be the first 
  // part of the original edge, the return value will be the second
  // part 
  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //   !!!   NOTE: THIS AFFECTS rev() IN THE OPPOSITE WAY:   !!!
  //   !!!   the origin of rev() will no longer be what it   !!!
  //   !!!   used to be; rev() will now start from the new   !!!
  //   !!!   vertex                                          !!!
  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  HalfEdge<V,F>* insertVertex(V* vertex = NULL);

  // insert a vertex into the face to the left of the current edge
  // and connect the new vertex to all vertices of the face in which
  // it was inserted; the return value is the edge from the new vertex
  // to the origin of the current vertex
  HalfEdge<V,F>* insertStar(V* vertex = NULL);

  // flip an edge in a triangulation (rotate it to the left):
  void flip();

  // starting from the common origin of e and f, cut from the face
  // to the left of e into the face to the left of f, towards the
  // given vertex. Return the edge from the new vertex to e
  static HalfEdge<V,F>* cutIntoFace(
    HalfEdge<V,F>* e, HalfEdge<V,F>* f, V* vertex = NULL);

  // collapse current edge and its predecessor on the face to the
  // left into one edge, and recurse on the successor, until the 
  // face is gone
  void zipUp();
};


//////////////////////////////////////////////////////////////////////
// one more traversal operator:                                     //
//////////////////////////////////////////////////////////////////////

// operator to proceed to sccorgn:
template <class V, class F>
HalfEdge<V,F>*& tosccorgn(HalfEdge<V,F>*& e);


//////////////////////////////////////////////////////////////////////
// save a little time for indexing other half-edges of a block:     //
//////////////////////////////////////////////////////////////////////

// when i is the index of a half-edge e in its pack of four, then 
// quadIndex[i][j] gives the index, relative to i, of the half-edge j,
// where j = 1 denotes rtol(e), j = 2 rev(e), j = 3 ltor(e):
const short int quadIndex[4][4] = {
  {0, 1, 2, 3}, {0, 1, 2, -1}, {0, 1, -2, -1}, {0, -3, -2, -1}
};


/**********************************************************************/

// IMPLEMENTATION OF HalfEdge

/**********************************************************************/


  int nrHalfEdgesAlive = 0;


  template <class V, class F>
  int HalfEdge<V,F>::totalNrAlive() {
    return nrHalfEdgesAlive;
  }

  template <class V, class F>
  void HalfEdge<V,F>::mark() {
    marking = true;
  }


  template <class V, class F>
  void HalfEdge<V,F>::unmark() {
    marking = false;
  }


  template <class V, class F>
  bool HalfEdge<V,F>::marked() const {
    return marking;
  }


  //////////////////////////////////////////////////////////////////////
  // public access to the half-edge and its neighbourhood:            //
  //////////////////////////////////////////////////////////////////////

  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::rev() {
    return this + quadIndex[halfEdgeIndex][2];
  }
  template <class V, class F>
  HalfEdge<F,V>* HalfEdge<V,F>::rtol() {
    return (HalfEdge<F,V>*) this + quadIndex[halfEdgeIndex][1];
  }
  template <class V, class F>
  HalfEdge<F,V>* HalfEdge<V,F>::ltor() {
    return (HalfEdge<F,V>*) this + quadIndex[halfEdgeIndex][3];
  }
  template <class V, class F>
  V* HalfEdge<V,F>::origin() {
    return (V*) pOrigin;
  }
  template <class V, class F>
  V* HalfEdge<V,F>::destin() {
    return rev()->origin();
  }
  template <class V, class F>
  F* HalfEdge<V,F>::leftFace() {
    return ltor()->origin();
  }
  template <class V, class F>
  F* HalfEdge<V,F>::rightFace() {
    return rtol()->origin();
  }
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::sccorgn() {
    return (HalfEdge<V,F>*) pSuccessor;
  } 
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::pccorgn() {
    return rtol()->sccorgn()->rtol();
  }
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::sccleft() {
    return ltor()->sccorgn()->rtol();
  }
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::pccleft() {
    return sccorgn()->rev();
  }
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::scright() {
    return rev()->sccorgn();
  }
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::pcright() {
    return rtol()->sccorgn()->ltor();
  }
  template <class V, class F>
  HalfEdge<F,V>* HalfEdge<V,F>::partner() {
    return sccorgn()->rtol();
  }


  //////////////////////////////////////////////////////////////////////
  // traversal operators:                                             //
  //////////////////////////////////////////////////////////////////////

  // collect all edges 
  // (one out of each primal half-edge and its reverse):
  template <class V, class F>
  std::vector<HalfEdge<V,F>* > HalfEdge<V,F>::getEdges()
  {
    // collect all edges
    std::vector<HalfEdge<V,F>* > toReport;
    int nrReady = 0;
    mark(); rev()->mark();
    toReport.push_back(this);
    do
    {
      HalfEdge<V,F>* f = toReport[nrReady]->sccorgn();
      HalfEdge<V,F>* g = toReport[nrReady]->rev()->sccorgn();
      if (!f->marked()) { f->mark(); f->rev()->mark(); toReport.push_back(f); }
      if (!g->marked()) { g->mark(); g->rev()->mark(); toReport.push_back(g); }
    } while (++nrReady < toReport.size());

    // clean up the marks
    for (
      typename std::vector<HalfEdge<V,F>* >::iterator i = toReport.begin(); 
      i != toReport.end(); 
      ++i
    ) 
    {
      (*i)->unmark(); (*i)->rev()->unmark();
    }

    // return the result
    return toReport;
  }

  // collect all vertices (one outgoing edge for each vertex)
  template <class V, class F>
  std::vector<HalfEdge<V,F>* > HalfEdge<V,F>::getVertices()
  {
    // collect all vertices
    std::vector<HalfEdge<V,F>* > toReport;
    std::stack<HalfEdge<V,F>* > toExplore;
    toExplore.push(this);
    do
    {
      HalfEdge<V,F>* start = toExplore.top(); toExplore.pop();

      // check if origin was reported already
      if (start->marked()) continue; 

      toReport.push_back(start);
      // mark all other edges with the same origin; those who receive
      // their first mark are put on the stack for reporting their other
      // end:
      HalfEdge<V,F>* e = start;
      do
      {
        e->mark();
        if (!e->rev()->marked()) toExplore.push(e->rev());
      } while (tosccorgn(e) != start);
    } while (!toExplore.empty());

    // clean up the marks
    for (
      typename std::vector<HalfEdge<V,F>* >::iterator i = toReport.begin(); 
      i != toReport.end(); 
      ++i
    )
    {
      HalfEdge<V,F>* e = *i;
      do e->unmark(); while (tosccorgn(e) != *i);
    }

    // return the result
    return toReport;
  }


  //////////////////////////////////////////////////////////////////////
  // counting and verification operators:                             //
  //////////////////////////////////////////////////////////////////////

  // get number of edges incident on start(e) (loops count twice):
  template <class V, class F>
  int HalfEdge<V,F>::nrEdgesOnOrigin() {
    HalfEdge<V,F>* e = this;
    int count = 0;
    do ++count; while (tosccorgn(e) != this);
    return count;
  }

  // get the complexity of the boundary of the face to the left of e:
  template <class V, class F>
  int HalfEdge<V,F>::nrEdgesOnLeftFace() {
    return ltor()->nrEdgesOnOrigin();
  }

  // check whether *this and *e have the same origin. Note: this follows
  // the cycles of edge around the origin; it does not rely on the 
  // pointers to vertex information! The loop follows the cycles around
  // both origins simultaneously and stops, at the latest, when the 
  // smallest of them has been traversed completely.
  template <class V, class F>
  bool HalfEdge<V,F>::hasSameOrigin(HalfEdge<V,F>* e) {
    HalfEdge<V,F>* walkAroundThis = this;
    HalfEdge<V,F>* walkAroundOther = e;
    do 
      if (walkAroundThis == e || walkAroundOther == this) 
        return true; 
    while (tosccorgn(walkAroundThis) != this && tosccorgn(walkAroundOther) != e);
    return false;
  }

  // check whether *this and e have the same face to the left. Note: 
  // this follows the cycle of edges around the face; it does not rely 
  // on pointers to face information!
  template <class V, class F>
  bool HalfEdge<V,F>::hasSameLeftFace(HalfEdge<V,F>* e) {
    return ltor()->hasSameOrigin(e->ltor());
  }


  //////////////////////////////////////////////////////////////////////
  // operations to create a graph                                     //
  //////////////////////////////////////////////////////////////////////

  // private constructor and post-constructor for individual halfedges:
  template <class V, class F>
  HalfEdge<V,F>::HalfEdge():
    marking(false)
  {
    ++nrHalfEdgesAlive;
  }

  template <class V, class F>
  void HalfEdge<V,F>::setData(
    int _halfEdgeIndex, void* _pOrigin, void* _pSuccessor)
  {
    halfEdgeIndex = _halfEdgeIndex;
    pOrigin = _pOrigin;
    pSuccessor = _pSuccessor;
  }

  // create a graph that consists of an edge from origin to 
  // destination with the given face as the outer and only face:
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::createEdge(
    V* origin, V* destination, F* face) 
  {
    HalfEdge<V,F>* e = new HalfEdge<V,F>[4];
    e[0].setData(0, (void*) origin, (void*) &e[0]);
    e[1].setData(1, (void*) face, (void*) &e[3]);
    e[2].setData(2, (void*) destination, (void*) &e[2]);
    e[3].setData(3, (void*) face, (void*) &e[1]);
    return e;
  }

  // create a graph that consists of a loop with the specified inner
  // face, outer face, and vertex; the half-edge that is returned has 
  // the inner face on the left:
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::createLoop(
    F* innerFace, V* vertex, F* outerFace) 
  {
    // implemented through createEdge on the dual:
    return HalfEdge<F,V>::createEdge(
      innerFace, outerFace, vertex
    )->rtol();
  }


  //////////////////////////////////////////////////////////////////////
  // operations to modify a graph                                     //
  //////////////////////////////////////////////////////////////////////

  // splice swaps sccorgn(e1) and sccorgn(e2), as well as 
  // sccorgn(partner(e1)) and sccorgn(partner(e2));
  // both swaps change whether the operands are in the same cycle around
  // a common origin (from yes to no or from no to yes);
  // splice is used by almost all methods that modify the graph;
  // we keep it private because it could have strange effects if used 
  // on pairs of edges it should not be used on
  template <class V, class F>
  void HalfEdge<V,F>::splice(HalfEdge<V,F>* e1, HalfEdge<V,F>* e2) {
    swap(
      e1->pSuccessor, 
      e2->pSuccessor
    );
    swap(
      e1->partner()->pSuccessor,
      e2->partner()->pSuccessor
    );
  }

  // the following four operations simply call splice, but after 
  // checking different preconditions that are neccesary for the 
  // intended effect to happen:

  // cut the common origin of e and f into two vertices, one with 
  // edges sccorgn(e),...,f and one with edges sccorgn(f),...,e;
  // thus merging the faces to the left of e and f.
  // (not to be used if it would render the graph disconnected)
  template <class V, class F>
  void HalfEdge<V,F>::splitOrigin(HalfEdge<V,F>* e, HalfEdge<V,F>* f)
  {
#ifdef SAFEQEDS
    assert(e->hasSameOrigin(f));
    // not to be used if it would render the graph disconnected:
    assert(!e->hasSameLeftFace(f));
#endif
#ifdef CONSISTENTQEDS
    // check whether the faces to be merged have the same label:
    assert(e->leftFace() == f->leftFace());
#endif
    
    // do the work:
    splice(e,f);
  }

  // the same as above when e and f have the same face to the left;
  // this will render the graph disconnected
  template <class V, class F>
  void HalfEdge<V,F>::splitGraph(HalfEdge<V,F>* e, HalfEdge<V,F>* f)
  {
#ifdef SAFEQEDS
    assert(e != f);
    assert(e->hasSameOrigin(f));
    assert(e->hasSameLeftFace(f));
#endif
    
    // do the work:
    splice(e,f);
  }

  // contract the origins of e and f into a single vertex, thus
  // cutting their common face into two faces
  template <class V, class F>
  void HalfEdge<V,F>::joinOrigins(HalfEdge<V,F>* e, HalfEdge<V,F>* f)
  {
    // implemented by calling SplitOrigin on the dual:
    HalfEdge<F,V>::splitOrigin(e->partner(), f->partner());
  }

  // contract the origins of the e and f into a single 
  // vertex to insert the graph with e on the outer face into the face 
  // to the left of f (or vice versa); 
  // we keep this one private, because it could have strange effects 
  // if used on pairs of edges it should not be used on, in particular,
  // if e and the current edge are in the same graph already.
  template <class V, class F>
  void HalfEdge<V,F>::insertGraph(HalfEdge<V,F>* e, HalfEdge<V,F>* f)
  {
#ifdef SAFEQEDS
    // really, e and f should not be in the same graph, but we do 
    // not want to traverse the whole graph to check
    assert(!e->hasSameOrigin(f));
    assert(!e->hasSameLeftFace(f));
#endif
#ifdef CONSISTENTQEDS
    // check whether the vertices and faces to be merged have the 
    // same label:
    assert(e->origin() == f->origin());
    assert(e->leftFace() == f->leftFace());
#endif

    // do the work:
    splice(e, f);
  }


  //////////////////////////////////////////////////////////////////////
  // operations to modify vertices and faces                          //
  //////////////////////////////////////////////////////////////////////

  // set vertex labels for all edges around the same vertex:
  template <class V, class F>
  void HalfEdge<V,F>::labelOrigin(V* v) {
    HalfEdge<V,F>* e = this;
    do e->pOrigin = (void*) v; while (tosccorgn(e) != this);
  }


  // set face labels for all edges around the face to the left:
  template <class V, class F>
  void HalfEdge<V,F>::labelLeftFace(F* f) {
    ltor()->labelOrigin(f);
  }


  //////////////////////////////////////////////////////////////////////
  // operations to delete (parts of) a graph                          //
  //////////////////////////////////////////////////////////////////////

  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //   !!!   NOTE: These operations do NOT delete the vertex !!!
  //   !!!   and face information that the vertex and face   !!!
  //   !!!   pointers point to!                              !!!
  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  template <class V, class F>
  HalfEdge<V,F>::~HalfEdge() {
    --nrHalfEdgesAlive;
  }

  // deleteEdge disconnects the edge from the rest of the graph and 
  // destroys it. we keep it private because it may render the rest of 
  // the graph disconnected, and we want to make sure that it is 
  // used only by operations that deal with that correctly
  template <class V, class F>
  void HalfEdge<V,F>::deleteEdge(HalfEdge<V,F>*& e) {
    // disconnect the edge from the rest of the graph
    splice(e, e->pccorgn());
    splice(e->rev(), e->sccleft()); 

    // destroy edge
    delete [] (e - e->halfEdgeIndex);
    e = NULL;
  }

  // deleteGraph deletes an entire graph
  template <class V, class F>
  void HalfEdge<V,F>::deleteGraph(HalfEdge<V,F>*& e)
  { 
    // one could simply uses getEdges and deleteEdge, but then we 
    // would waste (some) time on cleaning up marks and disconnecting
    // edges that are all going be thrown away anyway;
    // so let's do this more directly:

    // collect all edges:
    std::vector<HalfEdge<V,F>* > toDelete;
    int nrReady = 0;
    e->mark(); e->rev()->mark();
    toDelete.push_back(e);
    do
    {
      HalfEdge<V,F>* f = toDelete[nrReady]->sccorgn();
      HalfEdge<V,F>* g = toDelete[nrReady]->rev()->sccorgn();
      if (!f->marked()) { f->mark(); f->rev()->mark(); toDelete.push_back(f); }
      if (!g->marked()) { g->mark(); g->rev()->mark(); toDelete.push_back(g); }
    } while (++nrReady < toDelete.size());

    // throw them all away:
    for (
      typename std::vector<HalfEdge<V,F>* >::iterator i = toDelete.begin(); 
      i != toDelete.end(); 
      ++i
    ) 
      delete [] (*i - (*i)->halfEdgeIndex);

    e = NULL;
  }

  // deleteSubgraph removes the component that is surrounded, in
  // clockwise order, by the edges from e to, but not including, the 
  // current edge
  template <class V, class F>
  void HalfEdge<V,F>::deleteSubgraph(HalfEdge<V,F>*& e)
  {
    splitGraph(this, e);
    deleteGraph(e);
  }


  //////////////////////////////////////////////////////////////////////
  // advanced operations                                              //
  //////////////////////////////////////////////////////////////////////

  // cut the common origin of e and f into two vertices, one with 
  // edges sccorgn(e),...,f and one with edges sccorgn(f),...,e;
  // with a new edge connecting the two vertices; returns the new
  // edge in the direction from the origin of e to the origin of f
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::expandOrigin(HalfEdge<V,F>* e, HalfEdge<V,F>* f)
  {
    HalfEdge<V,F>* newEdge = 
      createLoop(f->leftFace(), e->origin(), e->leftFace());
    insertGraph(e, newEdge->rev());
    splitOrigin(newEdge, f);
    return newEdge;
  }

  // insert an edge between the origins of e and f, thus splitting 
  // their common face; returns the new edge in the direction from the 
  // origin of e to the origin of f
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::splitFace(HalfEdge<V,F>* e, HalfEdge<V,F>* f)
  {
    // implemented through expandOrigin on the dual:
    return HalfEdge<F,V>::expandOrigin(e->partner(), f->partner())->rtol();
  }

  // add a vertex on an edge; afterwards *this will be the first 
  // part of the original edge, the return value will be the second
  // part 
  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //   !!!   NOTE: THIS AFFECTS rev() IN THE OPPOSITE WAY:   !!!
  //   !!!   the origin of rev() will no longer be what it   !!!
  //   !!!   used to be; rev() will now start from the new   !!!
  //   !!!   vertex                                          !!!
  //   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::insertVertex(V* vertex)
  {
    expandOrigin(rev(), sccleft());
    rev()->labelOrigin(vertex);
    return rev()->sccorgn();
  }

  // insert a vertex into the face to the left of the current edge
  // and connect the new vertex to all vertices of the face in which
  // it was inserted; the return value is the edge from the new vertex
  // to the origin of the current vertex
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::insertStar(V* vertex)
  {
    HalfEdge<F,V>* edual = ltor()->sccorgn();
    HalfEdge<V,F>* newEdge = 
      createEdge(vertex, edual->rightFace(), leftFace());
    insertGraph(edual->rtol(), newEdge->rev());
    while (edual != ltor()) newEdge = splitFace(newEdge, tosccorgn(edual)->rtol());
    return newEdge;
  }

  // flip an edge in a triangulation (rotate it to the left):
  template <class V, class F>
  void HalfEdge<V,F>::flip()
  {
#ifdef SAFEQEDS
    assert(nrEdgesOnLeftFace() == 3);
    assert(rev()->nrEdgesOnLeftFace() == 3);
    assert(!hasSameLeftFace(rev()));
#endif
#ifdef CONSISTENTQEDS
    assert(leftFace() == rightFace());
#endif
    // get the quadrilateral
    HalfEdge<V,F>* q[4];
    q[0] = rev()->sccleft();
    q[1] = q[0]->sccleft();
    q[2] = sccleft();
    q[3] = q[2]->sccleft();
    // disconnect from q[0] and q[2]:
    splice(this, q[0]);
    splice(rev(), q[2]);
    labelOrigin(q[1]->origin());
    rev()->labelOrigin(q[3]->origin());
    // connect to q[1] and q[3]:
    splice(this, q[1]);
    splice(rev(), q[3]);
  }

  // starting from the common origin of e and f, cut from the face
  // to the left of e into the face to the left of f, towards the
  // given vertex. Return the edge from the new vertex to e.
  template <class V, class F>
  HalfEdge<V,F>* HalfEdge<V,F>::cutIntoFace(
    HalfEdge<V,F>* e, HalfEdge<V,F>* f, V* vertex)
  {
    return expandOrigin(e,f)->rev()->insertVertex(vertex);
  }

  // collapse current edge and its predecessor on the face to the
  // left into one edge, and recurse on the successor, until the 
  // face is gone
  template <class V, class F>
  void HalfEdge<V,F>::zipUp()
  {
    HalfEdge<V,F>* e = this;
    HalfEdge<V,F>* d = pccleft();
    while (true)
    {
      // check if the face is not a single-edge face (loop):
      assert(e != d); 
      HalfEdge<V,F>* f = e->sccleft();
      // check if the face is a two-edge face (final step):
      if (d == f) break;
      joinOrigins(d, f);
      d->labelLeftFace(d->rightFace());
      deleteEdge(d);
      e = f;
      d = e->pccleft();
    }
    d->labelLeftFace(d->rightFace());
    deleteEdge(d);
  }


  // to proceed to sccorgn:
  template <class V, class F>
  HalfEdge<V,F>*& tosccorgn(HalfEdge<V,F>*& e) 
  {
    e = e->sccorgn();
    return e;
  }


/**********************************************************************/

} // closes namespace AGBQEDS

#endif


/***********************************************************************

***********************************************************************/
